# SPDX-FileCopyrightText: Copyright (c) 2023-2025, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
import copy
import datetime
import inspect
import pathlib
import platform
import tempfile
import time
from contextlib import suppress

import anyio
import httpx
import pytest
import yaml

import kr8s
from kr8s._async_utils import anext
from kr8s._exceptions import NotFoundError
from kr8s._exec import CompletedExec, ExecError
from kr8s.asyncio.objects import (
    APIObject,
    ConfigMap,
    Deployment,
    Ingress,
    Node,
    PersistentVolume,
    Pod,
    Secret,
    Service,
    get_class,
    new_class,
    object_from_name_type,
    object_from_spec,
    objects_from_files,
)
from kr8s.asyncio.portforward import PortForward
from kr8s.objects import APIObject as SyncAPIObject
from kr8s.objects import Pod as SyncPod
from kr8s.objects import Service as SyncService
from kr8s.objects import (
    get_class as sync_get_class,
)
from kr8s.objects import (
    new_class as sync_new_class,
)
from kr8s.objects import (
    object_from_spec as sync_object_from_spec,
)
from kr8s.objects import objects_from_files as sync_objects_from_files

DEFAULT_TIMEOUT = httpx.Timeout(30)
CURRENT_DIR = pathlib.Path(__file__).parent


@pytest.fixture
async def nginx_pod(k8s_cluster, example_pod_spec):
    example_pod_spec["metadata"]["name"] = (
        "nginx-" + example_pod_spec["metadata"]["name"]
    )
    example_pod_spec["spec"]["containers"][0]["image"] = "nginx:latest"
    example_pod_spec["spec"]["containers"][0]["ports"] = [{"containerPort": 80}]
    example_pod_spec["spec"]["containers"][0]["readinessProbe"] = {
        "httpGet": {"path": "/", "port": 80},
        "initialDelaySeconds": 0,
        "periodSeconds": 1,
        "timeoutSeconds": 1,
        "successThreshold": 2,
    }
    example_pod_spec["metadata"]["labels"]["app"] = example_pod_spec["metadata"]["name"]
    pod = await Pod(example_pod_spec)
    await pod.create()
    while not await pod.ready():
        await anyio.sleep(0.1)
    await pod.exec(
        [
            "dd",
            "if=/dev/random",
            "of=/usr/share/nginx/html/foo.dat",
            "bs=4M",
            "count=10",
        ]
    )
    yield pod
    try:
        await pod.delete()
    except kr8s.NotFoundError:
        pass


@pytest.fixture
async def ubuntu_pod(k8s_cluster, example_pod_spec, ns):
    example_pod_spec["spec"]["containers"][0]["name"] = "ubuntu"
    example_pod_spec["spec"]["containers"][0]["image"] = "ubuntu:latest"
    example_pod_spec["spec"]["containers"][0]["command"] = ["sleep", "3600"]
    pod = await Pod(example_pod_spec)
    await pod.create()
    while not await pod.ready():
        await anyio.sleep(0.1)
    yield pod
    await pod.delete()


@pytest.fixture
async def nginx_service(example_service_spec, nginx_pod):
    example_service_spec["metadata"]["name"] = nginx_pod.name
    example_service_spec["spec"]["selector"] = nginx_pod.labels
    service = await Service(example_service_spec)
    await service.create()
    while not await service.ready():
        await anyio.sleep(0.1)  # pragma: no cover
    yield service
    try:
        await service.delete()
    except kr8s.NotFoundError:
        pass


async def test_pod_create_and_delete(example_pod_spec):
    pod = await Pod(example_pod_spec)
    await pod.create()
    with pytest.raises(NotImplementedError):
        pod.replicas
    assert await pod.exists()
    while not await pod.ready():
        await anyio.sleep(0.1)
    await pod.delete()
    while await pod.exists():
        await anyio.sleep(0.1)
    assert not await pod.exists()


async def test_pod_force_delete(example_pod_spec):
    pod = await Pod(example_pod_spec)
    await pod.create()
    while not await pod.exists():
        await anyio.sleep(0.1)
    await pod.delete(force=True)
    await anyio.sleep(0.1)
    assert not await pod.exists()


def test_pod_force_delete_sync(example_pod_spec):
    pod = SyncPod(example_pod_spec)
    pod.create()
    while not pod.exists():
        time.sleep(0.1)
    pod.delete(force=True)
    time.sleep(0.1)
    assert not pod.exists()


async def test_pod_object_from_name_type(example_pod_spec):
    pod = await Pod(example_pod_spec)
    await pod.create()
    assert await pod.exists()
    pod2 = await object_from_name_type(f"pod/{pod.name}", namespace=pod.namespace)
    pod3 = await object_from_name_type(f"pod.v1/{pod.name}", namespace=pod.namespace)
    assert pod2.name == pod.name
    assert type(pod2) is type(pod)
    assert pod3.name == pod.name
    assert type(pod3) is type(pod)
    await pod.delete()


async def test_pod_wait_ready(example_pod_spec):
    pod = await Pod(example_pod_spec)
    await pod.create()
    await pod.wait("condition=Ready")
    with pytest.raises(TimeoutError):
        await pod.wait("jsonpath='{.status.phase}'=Foo", timeout=0.1)
    await pod.wait("condition=Ready=true")
    await pod.wait("condition=Ready=True")
    await pod.wait("jsonpath='{.status.phase}'=Running")
    with pytest.raises(ValueError):
        await pod.wait("foo=NotARealCondition")
    await pod.delete(grace_period=10)
    await pod.wait("condition=Ready=False")
    await pod.wait("delete")


async def test_pod_missing_await_error(example_pod_spec):
    pod = Pod(example_pod_spec)  # We intentionally forget to await here
    assert pod._api is None
    with pytest.raises(RuntimeError, match="forget to await it"):
        await pod.create()


async def test_pod_wait_multiple_conditions(example_pod_spec):
    pod = await Pod(example_pod_spec)
    await pod.create()
    await pod.wait(conditions=["condition=Failed", "condition=Ready"])
    with pytest.raises(TimeoutError):
        await pod.wait(
            conditions=["condition=Failed", "condition=Ready"], mode="all", timeout=0.1
        )
    await pod.wait(
        conditions=[
            "condition=Initialized",
            "condition=ContainersReady",
        ],
        mode="all",
    )
    with pytest.raises(ValueError):
        await pod.wait(conditions=["condition=Failed", "condition=Ready"], mode="foo")
    await pod.delete()


def test_pod_wait_ready_sync(example_pod_spec):
    pod = SyncPod(example_pod_spec)
    pod.create()
    pod.wait("condition=Ready")
    with pytest.raises(TimeoutError):
        pod.wait("jsonpath='{.status.phase}'=Foo", timeout=0.1)
    pod.wait("condition=Ready=true")
    pod.wait("condition=Ready=True")
    pod.wait("jsonpath='{.status.phase}'=Running")
    with pytest.raises(ValueError):
        pod.wait("foo=NotARealCondition")
    pod.delete()
    pod.wait("condition=Ready=False")
    pod.wait("delete")


def test_wait_replicas(ns):
    from kr8s.objects import StatefulSet

    ss = StatefulSet(
        {
            "metadata": {
                "name": "test-wait-replicas",
                "namespace": ns,
            },
            "spec": {
                "replicas": 3,
                "selector": {
                    "matchLabels": {
                        "app": "test-wait-replicas",
                    }
                },
                "template": {
                    "metadata": {
                        "labels": {
                            "app": "test-wait-replicas",
                        }
                    },
                    "spec": {
                        "containers": [
                            {
                                "name": "test-wait-replicas",
                                "image": "nginx:latest",
                            }
                        ]
                    },
                },
            },
        }
    )
    ss.create()
    try:
        ss.wait("jsonpath='{.status.availableReplicas}'=3", timeout=30)
    finally:
        ss.delete()


def test_pod_refresh_sync(example_pod_spec):
    pod = SyncPod(example_pod_spec)
    pod.create()
    pod.refresh()
    pod.delete()


def test_pod_create_and_delete_sync(example_pod_spec):
    pod = SyncPod(example_pod_spec)
    pod.create()
    with pytest.raises(NotImplementedError):
        pod.replicas
    assert pod.exists()
    while not pod.ready():
        time.sleep(0.1)
    pod.delete()
    while pod.exists():
        time.sleep(0.1)
    assert not pod.exists()


async def test_list_and_ensure():
    api = await kr8s.asyncio.api()
    pods = [pod async for pod in api.get("pods", namespace=kr8s.ALL)]
    assert len(pods) > 0
    for pod in pods:
        await pod.refresh()
        assert await pod.exists(ensure=True)


async def test_nonexistant():
    pod = await Pod(
        {
            "apiVersion": "v1",
            "kind": "Pod",
            "metadata": {
                "name": "nonexistant",
                "namespace": "nonexistant",
            },
        }
    )
    assert not await pod.exists()
    with pytest.raises(kr8s.NotFoundError):
        await pod.exists(ensure=True)


async def test_pod_kind_api_raw():
    pod = await Pod(
        {
            "metadata": {"name": "foo"},
            "spec": {"containers": [{"name": "foo", "image": "nginx"}]},
        }
    )
    assert "kind" in pod.raw
    assert "apiVersion" in pod.raw
    assert pod.raw["kind"] == "Pod"
    assert pod.raw["apiVersion"] == "v1"


async def test_pod_metadata(example_pod_spec, ns):
    pod = await Pod(example_pod_spec)
    await pod.create()
    assert "name" in pod.metadata
    assert "hello" in pod.labels
    assert "foo" in pod.annotations
    assert ns == pod.namespace
    assert "example-" in pod.name
    assert "containers" in pod.spec
    assert "phase" in pod.status
    await pod.delete()


async def test_pod_missing_labels_annotations(example_pod_spec):
    del example_pod_spec["metadata"]["labels"]
    del example_pod_spec["metadata"]["annotations"]
    pod = await Pod(example_pod_spec)
    await pod.create()
    assert not pod.labels
    assert not pod.annotations
    await pod.delete()


async def test_pod_get(example_pod_spec):
    pod = await Pod(example_pod_spec)
    await pod.create()
    with pytest.raises(kr8s.NotFoundError):
        await Pod.get(f"{pod.name}-foo", namespace=pod.namespace, timeout=0.1)
    pod2 = await Pod.get(pod.name, namespace=pod.namespace)
    assert pod2.name == pod.name
    assert pod2.namespace == pod.namespace
    await pod.delete()
    while await pod.exists():
        await anyio.sleep(0.1)
    with pytest.raises(kr8s.NotFoundError):
        await pod2.delete()


def test_pod_get_sync(example_pod_spec):
    pod = SyncPod(example_pod_spec)
    pod.create()
    with pytest.raises(kr8s.NotFoundError):
        SyncPod.get(f"{pod.name}-foo", namespace=pod.namespace, timeout=0.1)
    pod2 = SyncPod.get(pod.name, namespace=pod.namespace)
    assert pod2.name == pod.name
    assert pod2.namespace == pod.namespace
    pod.delete()
    while pod.exists():
        time.sleep(0.1)
    with pytest.raises(kr8s.NotFoundError):
        pod2.delete()


@pytest.fixture
async def example_pod(example_pod_spec):
    pod = await Pod(example_pod_spec)
    await pod.create()
    yield pod
    await pod.delete()


async def test_pod_tolerate(example_pod):
    await example_pod.tolerate("key1", operator="Exists", effect="NoSchedule")
    await example_pod.tolerate(
        "key2",
        operator="Equal",
        effect="NoExecute",
        value="value1",
        toleration_seconds=600,
    )
    pod2 = await Pod.get(example_pod.name, namespace=example_pod.namespace)

    pod_objects_to_check = [example_pod, pod2]
    for pod in pod_objects_to_check:
        key_to_toleration = {}
        for toleration in pod.tolerations:
            if toleration["key"] in ("key1", "key2"):
                key_to_toleration[toleration["key"]] = toleration

        assert len(key_to_toleration) == 2
        assert key_to_toleration["key1"]["operator"] == "Exists"
        assert key_to_toleration["key1"]["effect"] == "NoSchedule"

        assert key_to_toleration["key2"]["operator"] == "Equal"
        assert key_to_toleration["key2"]["effect"] == "NoExecute"
        assert key_to_toleration["key2"]["value"] == "value1"
        assert key_to_toleration["key2"]["tolerationSeconds"] == 600


async def test_pod_from_name(example_pod_spec):
    pod = await Pod(example_pod_spec)
    await pod.create()
    pod2 = await Pod(
        pod.name, namespace=pod.namespace
    )  # Note: Does not call the Kubernetes API
    assert pod2.name == pod.name
    assert pod2.namespace == pod.namespace
    await pod.delete()
    while await pod.exists():
        await anyio.sleep(0.1)
    with pytest.raises(kr8s.NotFoundError):
        await pod2.delete()


async def test_pod_get_timeout(example_pod_spec):
    async def create_pod():
        await anyio.sleep(0.1)
        pod = await Pod(example_pod_spec)
        await pod.create()
        return pod

    async def get_pod():
        pod = await Pod.get(
            example_pod_spec["metadata"]["name"],
            namespace=example_pod_spec["metadata"]["namespace"],
            timeout=1,
        )
        return pod

    async with anyio.create_task_group() as tg:
        tg.start_soon(create_pod)
        tg.start_soon(get_pod)
    pod = await get_pod()
    await pod.delete()


async def test_missing_pod():
    with pytest.raises(kr8s.NotFoundError):
        await Pod.get("nonexistant", namespace="default")


@pytest.mark.parametrize("selector", ["abc=123def", {"abc": "123def"}])
async def test_label_selector(example_pod_spec, selector):
    example_pod_spec["metadata"]["labels"]["abc"] = "123def"
    pod = await Pod(example_pod_spec)
    await pod.create()

    api = await kr8s.asyncio.api()
    pods = [
        pod
        async for pod in api.get("pods", namespace=kr8s.ALL, label_selector=selector)
    ]
    assert len(pods) >= 0

    await pod.delete()


async def test_field_selector(example_pod_spec):
    pod = await Pod(example_pod_spec)
    await pod.create()

    api = await kr8s.asyncio.api()
    pods = [
        pod
        async for pod in api.get(
            "pods", namespace=kr8s.ALL, field_selector={"metadata.name": pod.name}
        )
    ]
    assert len(pods) == 1

    pods = [
        pod
        async for pod in api.get(
            "pods", namespace=kr8s.ALL, field_selector="metadata.name=" + "foo-bar-baz"
        )
    ]
    assert len(pods) == 0

    await pod.delete()


async def test_get_with_label_selector(example_pod_spec, ns):
    pod = await Pod(example_pod_spec)
    await pod.create()
    await pod.label(test="test_get_with_label_selector")

    pod2 = await Pod.get(label_selector=pod.labels, namespace=ns)
    assert pod == pod2

    pod3 = await Pod.get(field_selector={"metadata.name": pod.name}, namespace=ns)
    assert pod == pod3

    await pod.delete()


async def test_pod_watch(example_pod_spec):
    pod = await Pod(example_pod_spec)
    await pod.create()
    async for event, obj in pod.watch():
        assert event in ("ADDED", "MODIFIED", "DELETED")
        assert obj.name == pod.name
        break
    await pod.delete()


async def test_pod_annotate(example_pod_spec):
    pod = await Pod(example_pod_spec)
    await pod.create()
    await pod.annotate({"foo": "bar"})
    await pod.annotate(fizz="buzz")
    assert "foo" in pod.annotations
    assert "fizz" in pod.annotations
    with pytest.raises(ValueError):
        await pod.annotate({})
    await pod.delete()


async def test_pod_label(example_pod_spec):
    pod = await Pod(example_pod_spec)
    await pod.create()
    await pod.label({"foo": "bar"})
    assert "foo" in pod.labels
    with pytest.raises(ValueError):
        await pod.label({})
    await pod.label("foo-")
    assert "foo" not in pod.labels
    await pod.label(fizz="buzz")
    assert "fizz" in pod.labels
    await pod.remove_label("fizz")
    assert "fizz" not in pod.labels
    await pod.label({"foo/bar": "baz"})
    assert "foo/bar" in pod.labels
    await pod.label("foo/bar-")
    assert "foo/bar" not in pod.labels
    await pod.label(foo1="bar1", foo2="bar2")
    assert "foo1" in pod.labels
    assert "foo2" in pod.labels
    await pod.label("foo1-", "foo2-")
    assert "foo1" not in pod.labels
    assert "foo2" not in pod.labels
    await pod.delete()


def test_pod_label_sync(example_pod_spec):
    pod = SyncPod(example_pod_spec)
    pod.create()
    pod.label({"foo": "bar"})
    assert "foo" in pod.labels
    with pytest.raises(ValueError):
        pod.label({})
    pod.label("foo-")
    assert "foo" not in pod.labels
    pod.label(fizz="buzz")
    assert "fizz" in pod.labels
    pod.remove_label("fizz")
    assert "fizz" not in pod.labels
    pod.label({"foo/bar": "baz"})
    assert "foo/bar" in pod.labels
    pod.label("foo/bar-")
    assert "foo/bar" not in pod.labels
    pod.label(foo1="bar1", foo2="bar2")
    assert "foo1" in pod.labels
    assert "foo2" in pod.labels
    pod.label("foo1-", "foo2-")
    assert "foo1" not in pod.labels
    assert "foo2" not in pod.labels
    pod.delete()


def test_pod_watch_sync(example_pod_spec):
    pod = SyncPod(example_pod_spec)
    pod.create()
    for event, obj in pod.watch():
        assert event in ("ADDED", "MODIFIED", "DELETED")
        assert obj.name == pod.name
        break
    pod.delete()


async def test_patch_pod(example_pod_spec):
    pod = await Pod(example_pod_spec)
    await pod.create()
    assert "patched" not in pod.labels
    await pod.patch({"metadata": {"labels": {"patched": "true"}}})
    assert "patched" in pod.labels
    await pod.delete()


async def test_patch_pod_json(example_pod_spec):
    pod = await Pod(example_pod_spec)
    await pod.create()
    assert "patched" not in pod.labels
    await pod.patch(
        [{"op": "replace", "path": "/metadata/labels", "value": {"patched": "true"}}],
        type="json",
    )
    assert set(pod.labels) == {"patched"}
    await pod.delete()


async def test_all_v1_objects_represented():
    api = await kr8s.asyncio.api()
    k8s_objects = await api.api_resources()
    supported_apis = (
        "v1",
        "apps/v1",
        "autoscaling/v2",
        "batch/v1",
        "networking.k8s.io/v1",
        "policy/v1",
        "rbac.authorization.k8s.io/v1",
        "apiextensions.k8s.io/v1",
    )
    objects = [obj for obj in k8s_objects if obj["version"] in supported_apis]
    failures = []
    for obj in objects:
        try:
            assert get_class(obj["kind"], obj["version"])
        except KeyError:
            failures.append(f"{obj['kind']} ({obj['name']}.{obj['version']})")
    assert not failures, f"Failed to find {len(failures)} objects: {failures}"


async def test_object_from_spec(example_pod_spec, example_service_spec):
    pod = object_from_spec(example_pod_spec)
    assert isinstance(pod, Pod)
    assert pod.name == example_pod_spec["metadata"]["name"]
    assert pod.spec == example_pod_spec["spec"]
    assert pod._asyncio

    service = object_from_spec(example_service_spec)
    assert isinstance(service, Service)
    assert service.name == example_service_spec["metadata"]["name"]
    assert service.spec == example_service_spec["spec"]
    assert service._asyncio


async def test_object_from_spec_sync(example_pod_spec, example_service_spec):
    pod = sync_object_from_spec(example_pod_spec)
    assert isinstance(pod, Pod)
    assert pod.name == example_pod_spec["metadata"]["name"]
    assert pod.spec == example_pod_spec["spec"]
    assert not pod._asyncio

    service = sync_object_from_spec(example_service_spec)
    assert isinstance(service, Service)
    assert service.name == example_service_spec["metadata"]["name"]
    assert service.spec == example_service_spec["spec"]
    assert not service._asyncio


async def test_subclass_registration():
    with pytest.raises(KeyError):
        get_class("MyResource", "foo.kr8s.org/v1alpha1")

    class MyResource(APIObject):
        version = "foo.kr8s.org/v1alpha1"
        endpoint = "myresources"
        kind = "MyResource"
        plural = "myresources"
        singular = "myresource"
        namespaced = True

    get_class("MyResource", "foo.kr8s.org/v1alpha1")


async def test_new_class_registration():
    with pytest.raises(KeyError):
        get_class("MyOtherResource", "foo.kr8s.org/v1alpha1")

    MyOtherResource = new_class("MyOtherResource.foo.kr8s.org/v1alpha1")  # noqa: F841

    get_class("MyOtherResource", "foo.kr8s.org/v1alpha1")
    assert MyOtherResource._asyncio


async def test_new_sync_class_registration() -> None:
    with pytest.raises(KeyError):
        sync_get_class("MyOtherSyncResource", "foo.kr8s.org/v1alpha1")

    MyOtherSyncResource: type[SyncAPIObject] = sync_new_class(
        "MyOtherSyncResource.foo.kr8s.org/v1alpha1"
    )  # noqa: F841

    sync_get_class("MyOtherSyncResource", "foo.kr8s.org/v1alpha1")
    assert not MyOtherSyncResource._asyncio


async def test_new_class_registration_from_spec():
    my_async_resource_instance = await object_from_spec(
        {
            "kind": "MyAsyncResource",
            "apiVersion": "foo.kr8s.org/v1alpha1",
            "metadata": {"name": "foo"},
            "spec": {},
        },
        allow_unknown_type=True,
    )  # noqa: F841

    assert my_async_resource_instance._asyncio


async def test_new_sync_class_registration_from_spec():
    my_sync_resource_instance = sync_object_from_spec(
        {
            "kind": "MySyncResource",
            "apiVersion": "foo.kr8s.org/v1alpha1",
            "metadata": {"name": "foo"},
            "spec": {},
        },
        allow_unknown_type=True,
    )  # noqa: F841

    assert not my_sync_resource_instance._asyncio


async def test_class_registration_multiple_subclass():
    class MyResource(new_class("MyResource.foo.kr8s.org/v1alpha1")):
        def my_custom_method(self) -> str:
            return "foo"

    assert get_class("MyResource", "foo.kr8s.org/v1alpha1") is MyResource

    r = MyResource({})
    assert r.my_custom_method() == "foo"


async def test_deployment_scale(example_deployment_spec):
    deployment = await Deployment(example_deployment_spec)
    await deployment.create()
    assert deployment.replicas == 1
    await deployment.scale(2)
    assert deployment.replicas == 2
    while not await deployment.ready():
        await anyio.sleep(0.1)
    pods = await deployment.pods()
    assert len(pods) == 2
    await deployment.scale(1)
    assert deployment.replicas == 1
    await deployment.delete()


async def test_node():
    api = await kr8s.asyncio.api()
    nodes = [node async for node in api.get("nodes")]
    assert len(nodes) > 0
    for node in nodes:
        assert isinstance(node, Node)
        assert node.unschedulable is False
        await node.cordon()
        assert node.unschedulable is True
        await node.uncordon()


async def test_node_taint():
    api = await kr8s.asyncio.api()
    nodes = [node async for node in api.get("nodes")]
    assert len(nodes) > 0
    node = nodes[0]
    assert isinstance(node, Node)

    # Remove existing taints just in case they still exist
    for taint in node.taints:
        await node.taint(key=taint["key"], value=taint["value"], effect="NoSchedule-")
    assert not node.taints

    await node.taint(key="key1", value="value1", effect="NoSchedule")
    await node.taint(key="key2", value="value2", effect="NoSchedule")
    assert len(node.taints) == 2

    await node.taint(key="key1", value="value1", effect="NoSchedule-")
    await node.taint(key="key2", value="value2", effect="NoSchedule-")
    assert not node.taints

    with pytest.raises(NotFoundError):
        await node.taint(key="key123", value="value1", effect="NoSchedule-")


async def test_service_proxy():
    api = await kr8s.asyncio.api()
    service = await anext(api.get("services", "kubernetes"))
    assert isinstance(service, Service)
    assert service.name == "kubernetes"
    data = await service.proxy_http_get("/version", raise_for_status=False)
    assert isinstance(data, httpx.Response)


async def test_pod_logs(example_pod_spec):
    pod = await Pod(example_pod_spec)
    await pod.create()
    while not await pod.ready():
        await anyio.sleep(0.1)
    log = "\n".join([line async for line in pod.logs(container="pause")])
    assert isinstance(log, str)
    await pod.delete()


async def test_pod_port_forward_context_manager(nginx_service):
    [nginx_pod, *_] = await nginx_service.ready_pods()
    async with nginx_pod.portforward(80, local_port=None) as port:
        async with httpx.AsyncClient(timeout=DEFAULT_TIMEOUT) as session:
            resp = await session.get(f"http://localhost:{port}/")
            assert resp.status_code == 200
            resp = await session.get(f"http://localhost:{port}/foo")
            assert resp.status_code == 404
            resp = await session.get(f"http://localhost:{port}/foo.dat")
            assert resp.status_code == 200
            resp.read()


def test_pod_port_forward_context_manager_sync(nginx_service):
    nginx_service = SyncService.get(
        nginx_service.name, namespace=nginx_service.namespace
    )
    with nginx_service.portforward(80, local_port=None) as port:
        with httpx.Client(timeout=DEFAULT_TIMEOUT) as session:
            resp = session.get(f"http://localhost:{port}/")
            assert resp.status_code == 200
            resp = session.get(f"http://localhost:{port}/foo")
            assert resp.status_code == 404
            resp = session.get(f"http://localhost:{port}/foo.dat")
            assert resp.status_code == 200
            resp.read()


@pytest.mark.skip(reason="For manual testing only")
async def test_pod_port_forward_context_manager_manual(nginx_service):
    [nginx_pod, *_] = await nginx_service.ready_pods()
    pf = nginx_pod.portforward(80, 8184)
    async with pf:
        done = False
        while not done:
            # Put a breakpoint here and set done = True when you're finished.
            await anyio.sleep(1)


async def test_pod_port_forward_start_stop(nginx_service):
    [nginx_pod, *_] = await nginx_service.ready_pods()
    for _ in range(5):
        pf = nginx_pod.portforward(80, local_port=None)
        assert pf._bg_task is None
        port = await pf.start()
        assert pf._bg_task is not None
        async with httpx.AsyncClient(timeout=DEFAULT_TIMEOUT) as session:
            resp = await session.get(f"http://localhost:{port}/")
            assert resp.status_code == 200
            resp = await session.get(f"http://localhost:{port}/foo")
            assert resp.status_code == 404
            resp = await session.get(f"http://localhost:{port}/foo.dat")
            assert resp.status_code == 200
            resp.read()
        await pf.stop()
        assert pf._bg_task is None


async def test_service_port_forward_context_manager(nginx_service):
    async with nginx_service.portforward(80, local_port=None) as port:
        async with httpx.AsyncClient(timeout=DEFAULT_TIMEOUT) as session:
            resp = await session.get(f"http://localhost:{port}/")
            assert resp.status_code == 200
            resp = await session.get(f"http://localhost:{port}/foo")
            assert resp.status_code == 404


async def test_service_port_forward_start_stop(nginx_service):
    pf = nginx_service.portforward(80, local_port=None)
    assert pf._bg_task is None
    port = await pf.start()
    assert pf._bg_task is not None

    async with httpx.AsyncClient(timeout=DEFAULT_TIMEOUT) as session:
        resp = await session.get(f"http://localhost:{port}/")
        assert resp.status_code == 200
        resp = await session.get(f"http://localhost:{port}/foo")
        assert resp.status_code == 404

    await pf.stop()
    assert pf._bg_task is None


async def test_unsupported_port_forward():
    pv = await PersistentVolume({"metadata": {"name": "foo"}})
    with pytest.raises(AttributeError):
        await pv.portforward(80, local_port=None)  # type: ignore
    with pytest.raises(ValueError):
        await PortForward(pv, 80, local_port=None).start()


@pytest.mark.skipif(
    "macOS" in platform.platform(),
    reason="Hangs on macOS, see https://github.com/kr8s-org/kr8s/issues/380",
)
async def test_multiple_bind_addresses_port_forward(nginx_service):
    [nginx_pod, *_] = await nginx_service.ready_pods()

    # Example multiple addresses
    multiple_addresses = ["127.0.0.2", "127.0.0.3"]

    pf = nginx_pod.portforward(80, local_port=None, address=multiple_addresses)

    # Start the port forwarding
    await pf.start()
    async with httpx.AsyncClient(timeout=DEFAULT_TIMEOUT) as session:
        for address in multiple_addresses:
            resp = await session.get(f"http://{address}:{pf.local_port}/")
            assert resp.status_code == 200

    # Stop the port forwarding
    await pf.stop()


async def test_scalable_dot_notation():
    class Foo(APIObject):
        version = "foo.kr8s.org/v1alpha1"
        endpoint = "foos"
        kind = "Foo"
        plural = "foos"
        singular = "foo"
        namespaced = True
        scalable = True
        scalable_spec = "nested.replicas"

    foo = await Foo({"metadata": {"name": "foo"}, "spec": {"nested": {"replicas": 1}}})
    assert foo.replicas == 1


async def test_object_from_file():
    api = await kr8s.asyncio.api()
    objects = await objects_from_files(
        CURRENT_DIR / "resources" / "simple" / "nginx_pod.yaml", api=api
    )
    assert len(objects) == 1
    assert isinstance(objects[0], Pod)
    assert objects[0].kind == "Pod"
    assert objects[0].name == "nginx"
    assert len(objects[0].spec.containers) == 1


async def test_objects_from_file():
    objects = await objects_from_files(
        CURRENT_DIR / "resources" / "simple" / "nginx_pod_service.yaml"
    )
    assert len(objects) == 2
    assert isinstance(objects[0], Pod)
    assert isinstance(objects[1], Service)
    assert len(objects[1].spec.ports) == 1


def test_objects_from_file_sync():
    objects = list(
        sync_objects_from_files(
            CURRENT_DIR / "resources" / "simple" / "nginx_pod_service.yaml"
        )
    )
    assert len(objects) == 2
    assert isinstance(objects[0], Pod)
    assert isinstance(objects[1], Service)
    assert not objects[0]._asyncio
    assert objects[0].api
    assert not objects[0].api._asyncio
    assert len(objects[1].spec.ports) == 1


async def test_objects_from_files():
    simple_dir = CURRENT_DIR / "resources" / "simple"
    objects = await objects_from_files(simple_dir)
    assert len(objects) > 1


async def test_objects_from_files_nested():
    simple_dir = CURRENT_DIR / "resources" / "simple"

    objects = await objects_from_files(simple_dir)
    assert not any(isinstance(o, Ingress) for o in objects)

    objects = await objects_from_files(simple_dir, recursive=True)
    assert any(isinstance(o, Ingress) for o in objects)


async def test_custom_object_from_file():
    simple_dir = CURRENT_DIR / "resources" / "custom" / "evc.yaml"
    objects = await objects_from_files(simple_dir)
    assert len(objects) == 1


async def test_pod_to_dict(example_pod_spec):
    pod = Pod(example_pod_spec)
    assert dict(pod) == example_pod_spec
    assert dict(pod) == pod.raw


async def test_pod_to_yaml(example_pod_spec):
    pod = Pod(example_pod_spec)
    assert f"name: {pod.name}" in pod.to_yaml()
    assert yaml.safe_load(pod.to_yaml()) == example_pod_spec


async def test_adoption(nginx_service):
    [nginx_pod, *_] = await nginx_service.ready_pods()
    await nginx_service.adopt(nginx_pod)
    assert "ownerReferences" in nginx_pod.metadata
    assert nginx_pod.metadata["ownerReferences"][0]["name"] == nginx_service.name
    await nginx_service.delete()
    while await nginx_pod.exists():
        await anyio.sleep(0.1)


async def test_cast_to_from_lightkube(example_pod_spec):
    pytest.importorskip("lightkube")
    from lightkube import codecs
    from lightkube.resources.core_v1 import Pod as LightkubePod

    starting_pod = codecs.from_dict(example_pod_spec)

    kr8s_pod = await Pod(starting_pod)
    assert isinstance(kr8s_pod, Pod)
    assert kr8s_pod.name == example_pod_spec["metadata"]["name"]
    assert kr8s_pod.namespace == example_pod_spec["metadata"]["namespace"]
    assert kr8s_pod.kind == "Pod"
    assert kr8s_pod.version == "v1"

    lightkube_pod = kr8s_pod.to_lightkube()
    assert isinstance(lightkube_pod, LightkubePod)
    assert lightkube_pod.metadata
    assert lightkube_pod.metadata.name == example_pod_spec["metadata"]["name"]
    assert lightkube_pod.metadata.namespace == example_pod_spec["metadata"]["namespace"]


async def test_cast_to_from_kubernetes(example_pod_spec):
    kubernetes = pytest.importorskip("kubernetes")

    starting_pod = kubernetes.client.models.v1_pod.V1Pod(
        api_version=example_pod_spec["apiVersion"],
        kind=example_pod_spec["kind"],
        metadata=example_pod_spec["metadata"],
        spec=example_pod_spec["spec"],
    )

    kr8s_pod = await Pod(starting_pod)
    assert isinstance(kr8s_pod, Pod)
    assert kr8s_pod.name == example_pod_spec["metadata"]["name"]
    assert kr8s_pod.namespace == example_pod_spec["metadata"]["namespace"]
    assert kr8s_pod.kind == "Pod"
    assert kr8s_pod.version == "v1"


async def test_cast_to_from_kubernetes_asyncio(example_pod_spec):
    kubernetes_asyncio = pytest.importorskip("kubernetes_asyncio")

    starting_pod = kubernetes_asyncio.client.models.v1_pod.V1Pod(
        api_version=example_pod_spec["apiVersion"],
        kind=example_pod_spec["kind"],
        metadata=example_pod_spec["metadata"],
        spec=example_pod_spec["spec"],
    )

    kr8s_pod = await Pod(starting_pod)
    assert isinstance(kr8s_pod, Pod)
    assert kr8s_pod.name == example_pod_spec["metadata"]["name"]
    assert kr8s_pod.namespace == example_pod_spec["metadata"]["namespace"]
    assert kr8s_pod.kind == "Pod"
    assert kr8s_pod.version == "v1"


async def test_cast_to_from_pykube_ng(example_pod_spec):
    pykube = pytest.importorskip("pykube")

    starting_pod = pykube.objects.Pod(None, example_pod_spec)

    kr8s_pod = await Pod(starting_pod)
    assert isinstance(kr8s_pod, Pod)
    assert kr8s_pod.name == example_pod_spec["metadata"]["name"]
    assert kr8s_pod.namespace == example_pod_spec["metadata"]["namespace"]
    assert kr8s_pod.kind == "Pod"
    assert kr8s_pod.version == "v1"

    pykube_pod = kr8s_pod.to_pykube(None)
    assert isinstance(pykube_pod, pykube.objects.Pod)
    assert pykube_pod.name == example_pod_spec["metadata"]["name"]
    assert pykube_pod.namespace == example_pod_spec["metadata"]["namespace"]


async def test_to_dict(example_pod_spec):
    pod = await Pod(example_pod_spec)
    to_spec = pod.to_dict()
    assert to_spec == example_pod_spec
    assert isinstance(to_spec, dict)


async def test_pod_exec(ubuntu_pod):
    ex = await ubuntu_pod.exec(["date"])
    assert isinstance(ex, CompletedExec)
    assert str(datetime.datetime.now().year) in ex.stdout.decode()
    assert ex.args == ["date"]
    assert ex.stderr == b""
    assert ex.returncode == 0


async def test_pod_exec_error(ubuntu_pod):
    with pytest.raises(ExecError):
        await ubuntu_pod.exec(["date", "foo"])

    ex = await ubuntu_pod.exec(["date", "foo"], check=False)
    assert ex.args == ["date", "foo"]
    assert b"invalid date" in ex.stderr
    assert ex.returncode == 1

    with pytest.raises(ExecError):
        ex.check_returncode()


async def test_pod_exec_to_file(ubuntu_pod):
    with tempfile.TemporaryFile(mode="w+b") as tmp:
        exc = await ubuntu_pod.exec(["date"], stdout=tmp, capture_output=False)
        tmp.seek(0)
        assert str(datetime.datetime.now().year) in tmp.read().decode()
        assert exc.stdout == b""

    with tempfile.TemporaryFile(mode="w+b") as tmp:
        with pytest.raises(ExecError):
            await ubuntu_pod.exec(["date", "foo"], stderr=tmp)
        tmp.seek(0)
        assert b"invalid date" in tmp.read()


@pytest.mark.xfail(reason="Exec protocol v5.channel.k8s.io not available")
async def test_pod_exec_stdin(ubuntu_pod):
    ex = await ubuntu_pod.exec(["cat"], stdin="foo")
    assert b"foo" in ex.stdout


async def test_pod_exec_not_ready(ns):
    pod = await Pod.gen(name="nginx", namespace=ns, image="nginx:latest")
    await pod.create()
    try:
        assert not await pod.ready()
        await pod.exec(["date"])
        assert await pod.ready()
    finally:
        await pod.delete()


async def test_service_exec(nginx_service):
    ex = await nginx_service.exec(["date"])
    assert isinstance(ex, CompletedExec)
    assert str(datetime.datetime.now().year) in ex.stdout.decode()
    assert ex.args == ["date"]
    assert ex.stderr == b""
    assert ex.returncode == 0


async def test_configmap_exec_raises():
    cm = await ConfigMap(
        {"metadata": {"name": "nginx", "namespace": "default"}, "data": {"foo": "bar"}}
    )
    with pytest.raises(NotImplementedError):
        await cm.exec(["date"])


async def test_configmap_data(ns):
    [cm] = await objects_from_files(CURRENT_DIR / "resources" / "configmap.yaml")
    assert isinstance(cm, ConfigMap)
    cm.namespace = ns
    await cm.create()
    assert "game.properties" in cm.data
    assert cm.data.player_initial_lives == "3"
    assert "color.good=purple" in cm.data["user-interface.properties"]
    await cm.delete()


async def test_secret_data(ns):
    [secret] = await objects_from_files(CURRENT_DIR / "resources" / "secret.yaml")
    assert isinstance(secret, Secret)
    secret.namespace = ns
    await secret.create()
    assert "tls.crt" in secret.data
    await secret.delete()


async def test_secret_create_delete_not_changed(ns):
    [secret] = await objects_from_files(CURRENT_DIR / "resources" / "secret.yaml")
    secret.namespace = ns
    await secret.create()
    await secret.delete()
    exists = await secret.exists()
    assert not exists


async def test_validate_pod(example_pod_spec):
    kubernetes_validate = pytest.importorskip("kubernetes_validate")
    pod = await Pod(example_pod_spec)
    kubernetes_validate.validate(pod.raw, "1.28", strict=True)


async def test_validate_pod_fail(bad_pod_spec):
    kubernetes_validate = pytest.importorskip("kubernetes_validate")
    pod = await Pod(bad_pod_spec)
    with pytest.raises(kubernetes_validate.ValidationError):
        kubernetes_validate.validate(pod.raw, "1.28", strict=True)


async def test_pod_errors(bad_pod_spec):
    pod = await Pod(bad_pod_spec)
    with pytest.raises(kr8s.ServerError, match="Required value"):
        await pod.create()


async def test_pod_list():
    pods1 = [pod async for pod in kr8s.asyncio.get("pods", namespace=kr8s.ALL)]
    pods2 = [pod async for pod in Pod.list(namespace=kr8s.ALL)]
    assert pods1 and pods2
    assert len(pods1) == len(pods2)
    assert all(isinstance(p, Pod) for p in pods1)
    assert all(isinstance(p, Pod) for p in pods2)
    assert {p.name for p in pods1} == {p.name for p in pods2}


def test_pod_list_sync():
    pods1 = [pod for pod in kr8s.get("pods", namespace=kr8s.ALL)]
    pods2 = [pod for pod in SyncPod.list(namespace=kr8s.ALL)]
    assert pods1 and pods2
    assert len(pods1) == len(pods2)
    assert all(isinstance(p, SyncPod) for p in pods1)
    assert all(isinstance(p, SyncPod) for p in pods2)
    assert {p.name for p in pods1} == {p.name for p in pods2}


async def test_pod_list_api():
    api = await kr8s.asyncio.api()
    pods = [pod async for pod in Pod.list(namespace=kr8s.ALL, api=api)]
    assert pods
    assert pods[0].api
    assert pods[0].api == api
    assert pods[0].api._asyncio


async def test_pod_list_api_sync():
    api = kr8s.api()
    pods = [pod for pod in SyncPod.list(namespace=kr8s.ALL, api=api)]
    assert pods
    assert pods[0].api
    assert pods[0].api == api
    assert not pods[0].api._asyncio


@pytest.mark.parametrize(
    "ports",
    [
        80,
        [80],
        [80, 81],
        [{"containerPort": 80}],
        [{"containerPort": 80}, {"containerPort": 81}],
    ],
)
async def test_pod_gen_ports(ns, ports):
    pod = await Pod.gen(name="nginx", namespace=ns, image="nginx:latest", ports=ports)
    try:
        await pod.create()  # This should succeed
    finally:
        with suppress(kr8s.NotFoundError):
            await pod.delete()


def test_sync_new_class_is_sync():
    MyResource = new_class(
        kind="MyResource",
        version="newclass.example.com/v1",
        namespaced=True,
        asyncio=False,
    )
    instance = MyResource({})
    assert not instance._asyncio
    assert not inspect.iscoroutinefunction(instance.create)


def test_new_class_plural_suffix():
    MyFoo = new_class(
        kind="MyFoo",
        version="newclass.example.com/v1",
        namespaced=True,
    )
    instance = MyFoo({})
    assert instance.plural == "myfoos"
    assert instance.endpoint == "myfoos"

    MyClass = new_class(
        kind="MyClass",
        plural="MyClasses",
        version="newclass.example.com/v1",
        namespaced=True,
    )
    instance = MyClass({})
    assert instance.plural == "myclasses"
    assert instance.endpoint == "myclasses"

    MyPolicy = new_class(
        kind="MyPolicy",
        plural="MyPolicies",
        version="newclass.example.com/v1",
        namespaced=True,
    )
    instance = MyPolicy({})
    assert instance.plural == "mypolicies"
    assert instance.endpoint == "mypolicies"


def test_object_setter(example_pod_spec):
    po = Pod(example_pod_spec)

    assert po.name != "foo"
    po.raw["metadata"]["name"] = "foo"
    assert po.name == "foo"

    assert po.raw["spec"]["containers"][0]["name"] != "bar"
    po.raw["spec"]["containers"][0]["name"] = "bar"
    assert po.raw["spec"]["containers"][0]["name"] == "bar"


def test_object_setter_from_old_spec(example_pod_spec):
    spec = copy.deepcopy(example_pod_spec)

    po = Pod(example_pod_spec)

    assert po.raw["spec"]["containers"][0]["name"] != "bar"
    po.raw["spec"]["containers"][0]["name"] = "bar"
    assert po.raw["spec"]["containers"][0]["name"] == "bar"

    new_po = Pod(spec)
    assert new_po.raw["spec"]["containers"][0]["name"] != "bar"
    new_po.raw["spec"] = po.raw["spec"]
    assert new_po.raw["spec"]["containers"][0]["name"] == "bar"


def test_parse_kind():
    from kr8s._objects import parse_kind

    assert parse_kind("Pod") == ("pod", "", "")
    assert parse_kind("Pods") == ("pods", "", "")
    assert parse_kind("pod/v1") == ("pod", "", "v1")
    assert parse_kind("deploy") == ("deploy", "", "")
    assert parse_kind("gateway") == ("gateway", "", "")
    assert parse_kind("gateways") == ("gateways", "", "")
    assert parse_kind("gateway.networking.istio.io") == (
        "gateway",
        "networking.istio.io",
        "",
    )
    assert parse_kind("gateways.networking.istio.io") == (
        "gateways",
        "networking.istio.io",
        "",
    )
    assert parse_kind("gateway.v1.networking.istio.io") == (
        "gateway",
        "networking.istio.io",
        "v1",
    )
    assert parse_kind("gateways.v1.networking.istio.io") == (
        "gateways",
        "networking.istio.io",
        "v1",
    )
    assert parse_kind("gateway.networking.istio.io/v1") == (
        "gateway",
        "networking.istio.io",
        "v1",
    )
    assert parse_kind("gateways.networking.istio.io/v1") == (
        "gateways",
        "networking.istio.io",
        "v1",
    )


async def test_setting_attributes():
    po = await Pod.gen(name="nginx", image="nginx:latest")
    po.metadata.labels = {"foo": "bar"}
    assert po.metadata.labels == {"foo": "bar"}

    po.metadata.generateName = po.metadata.pop("name") + "-"
    assert "generateName" in po.metadata
    assert "name" not in po.metadata

    po.name = "abc123"
    assert po.name == "abc123"
    po.namespace = "bar"
    assert po.namespace == "bar"
    po.metadata = {"name": "def", "namespace": "buzz"}
    assert po.name == "def"
    po["metadata"] = {"name": "ghi", "namespace": "buzz"}
    assert po.name == "ghi"

    po.spec.containers[0].image = "wordpress:latest"
    assert po.spec.containers[0].image == "wordpress:latest"

    with pytest.raises(NotImplementedError):
        po.replicas = 2


async def test_generate_name():
    po = await kr8s.asyncio.objects.Pod.gen(
        generate_name="nginx-", image="nginx:latest"
    )

    assert "generateName" in po.metadata
    assert po.metadata.generateName == "nginx-"
    assert "name" not in po.metadata
    with pytest.raises(ValueError):
        assert po.name

    assert "generateName(nginx-)" in po.__repr__()

    await po.create()
    try:
        assert po.name
        assert po.name.startswith("nginx-")
        assert len(po.name) > len(po.metadata.generateName)
        assert po.metadata.generateName in po.name
    finally:
        await po.delete()


async def test_recreate_pod(ns):
    po = await kr8s.asyncio.objects.Pod.gen(
        generate_name="nginx-", image="nginx:latest", namespace=ns
    )
    await po.create()
    await po.delete()
    while await po.exists():
        await anyio.sleep(0.1)
    await po.create()
    while not await po.exists():
        await anyio.sleep(0.1)
    assert await po.exists()
    await po.delete()


async def test_create_existing_pod_fails():
    po = await kr8s.asyncio.objects.Pod.gen(
        generate_name="nginx-", image="nginx:latest"
    )
    await po.create()
    with pytest.raises(kr8s.ServerError, match="already exists"):
        await po.create()



# SPDX-FileCopyrightText: Copyright (c) 2023-2025, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License

from datetime import datetime

import anyio
import rich.table
import typer
from rich import box
from rich.console import Console
from rich.live import Live
from rich.text import Text

import kr8s
from kr8s._async_utils import anext
from kr8s.asyncio.objects import Table

TIMESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
CHANGE_INDICATOR_DURATION = 15  # seconds to show change indicators
EXCLUDED_CHANGE_COLUMNS = ["Age"]  # columns to exclude from change indicators

# Unicode symbols for changes with their styles
UP_ARROW = Text("↑", style="green")
DOWN_ARROW = Text("↓", style="red")
CHANGE_DELTA = Text("Δ", style="cyan")


def is_numeric(value):
    """Check if a string value can be converted to a number."""
    try:
        float(value)
        return True
    except (ValueError, TypeError):
        return False


def get_change_indicator(old_value, new_value):
    """Get the appropriate change indicator based on value comparison."""
    if not is_numeric(old_value) or not is_numeric(new_value):
        return CHANGE_DELTA if old_value != new_value else Text("")

    old_num = float(old_value)
    new_num = float(new_value)

    if new_num > old_num:
        return UP_ARROW
    elif new_num < old_num:
        return DOWN_ARROW
    return Text("")


console = Console()


async def draw_table(
    kind, response, resource_names, previous_state=None, last_update_time=None
):
    table = rich.table.Table(box=box.SIMPLE)
    table.add_column("Namespace", style="magenta", no_wrap=True)

    for column in response.column_definitions:
        if column["priority"] == 0:
            kwargs = {}
            if column["name"] == "Name":
                kwargs = {"style": "cyan", "no_wrap": True}
            table.add_column(column["name"], **kwargs)

    current_time = datetime.now()
    show_indicators = (
        last_update_time
        and (current_time - last_update_time).total_seconds()
        < CHANGE_INDICATOR_DURATION
    )

    for row in response.rows:
        if not resource_names or row["object"]["metadata"]["name"] == resource_names[0]:
            cells = []
            for cell, column in zip(row["cells"], response.column_definitions):
                if column["priority"] == 0:
                    cell_str = str(cell)
                    if (
                        show_indicators
                        and previous_state
                        and column["name"] not in EXCLUDED_CHANGE_COLUMNS
                    ):
                        # Find matching previous row
                        prev_row = next(
                            (
                                r
                                for r in previous_state.rows
                                if r["object"]["metadata"]["name"]
                                == row["object"]["metadata"]["name"]
                                and r["object"]["metadata"]["namespace"]
                                == row["object"]["metadata"]["namespace"]
                            ),
                            None,
                        )
                        if prev_row:
                            prev_cell = next(
                                (
                                    c
                                    for c, col in zip(
                                        prev_row["cells"],
                                        previous_state.column_definitions,
                                    )
                                    if col["priority"] == 0
                                    and col["name"] == column["name"]
                                ),
                                None,
                            )
                            if prev_cell is not None:
                                indicator = get_change_indicator(
                                    str(prev_cell), cell_str
                                )
                                if indicator:
                                    cell_str = indicator + Text(" ") + Text(cell_str)
                    cells.append(cell_str)
            table.add_row(row["object"]["metadata"]["namespace"], *cells)

    if not table.rows:
        if resource_names:
            console.print(f"No resources found with name {resource_names[0]}.")
        else:
            console.print(f"No {kind} resources found in current namespace.")
        return None
    return table


async def get_resources(resources, label_selector, field_selector):
    data = {}
    for kind in resources:
        data[kind] = await anext(
            kr8s.asyncio.get(
                kind,
                label_selector=label_selector,
                field_selector=field_selector,
                as_object=Table,
            )
        )
    return data


async def get(
    resources: list[str] = typer.Argument(..., help="TYPE[.VERSION][.GROUP]"),
    all_namespaces: bool = typer.Option(
        False,
        "-A",
        "--all-namespaces",
        help="If present, list the requested object(s) across all namespaces. "
        "Namespace in current context is ignored even if specified with --namespace.",
    ),
    namespace: str = typer.Option(
        None,
        "-n",
        "--namespace",
    ),
    label_selector: str = typer.Option(
        "",
        "-l",
        "--selector",
        help="Selector (label query) to filter on, supports '=', '==', and '!='. "
        "(e.g. -l key1=value1,key2=value2). "
        "Matching objects must satisfy all of the specified label constraints.",
    ),
    field_selector: str = typer.Option(
        "",
        "--field-selector",
        help="Selector (field query) to filter on, supports '=', '==', and '!='. "
        "(e.g. --field-selector key1=value1,key2=value2). "
        "The server only supports a limited number of field queries per type. ",
    ),
    show_kind: bool = typer.Option(
        False,
        "--show-kind",
        help="If present, list the resource type for the requested object(s).",
    ),
    show_labels: bool = typer.Option(
        False,
        "--show-labels",
        help="When printing, show all labels as the last column "
        "(default hide labels column).",
    ),
    label_columns: list[str] = typer.Option(
        [],
        "-L",
        "--label-columns",
        help="Accepts a comma separated list of labels that are going to be presented as columns. "
        "Names are case-sensitive. "
        "You can also use multiple flag options like -L label1 -L label2...",
    ),
    watch: bool = typer.Option(
        False,
        "-w",
        "--watch",
        help="After listing/getting the requested object, watch for changes.",
    ),
):
    """Display one or many resources.

    Prints a table of the most important information about the specified resources. You can filter the list using a
    label selector and the --selector flag. If the desired resource type is namespaced you will only see results
    in your current namespace unless you pass --all-namespaces.

    By specifying the output as 'template' and providing a Jinja2 template as the value of the --template flag, you
    can filter the attributes of the fetched resources.

    Use "kubectl-ng api-resources" for a complete list of supported resources.
    """
    resource_names = []
    # Support kubectl-ng get pod,svc
    if len(resources) == 1:
        resources = resources[0].split(",")
        if len(resources) > 1 and watch:
            raise typer.BadParameter("you may only specify a single resource type")
    # Support kubectl-ng get pod [name]
    elif len(resources) == 2:
        resources, resource_names = ([r] for r in resources)
    else:
        raise typer.BadParameter(
            f"Format '{' '.join(resources)}' not currently supported."
        )
    kubernetes = await kr8s.asyncio.api()
    if namespace:
        kubernetes.namespace = namespace
    if all_namespaces:
        kubernetes.namespace = kr8s.ALL

    data = await get_resources(resources, label_selector, field_selector)

    if not watch:
        for kind, response in data.items():
            table = await draw_table(kind, response, resource_names)
            if table:
                console.print(table)
    else:
        kind = list(data)[0]
        response = data[kind]
        table = await draw_table(kind, response, resource_names)
        previous_state = None
        last_update_time = None

        with Live(table, console=console, auto_refresh=False) as live:
            while True:
                await anyio.sleep(5)
                previous_state = response
                last_update_time = datetime.now()
                data = await get_resources(resources, label_selector, field_selector)
                for kind, response in data.items():
                    table = await draw_table(
                        kind, response, resource_names, previous_state, last_update_time
                    )
                live.update(table)
                live.refresh()



# SPDX-FileCopyrightText: Copyright (c) 2023-2025, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
"""kubectl-ng."""

try:
    from .__version import version as __version__  # noqa
    from .__version import version_tuple as __version_tuple__  # noqa
except ImportError:
    __version__ = "0.0.0"
    __version_tuple__ = (0, 0, 0)



# SPDX-FileCopyrightText: Copyright (c) 2023-2025, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
"""
This module contains `kr8s`, a simple, extensible Python client library for Kubernetes.

At the top level, `kr8s` provides a synchronous API that wraps the asynchronous API provided by `kr8s.asyncio`.
Both APIs are functionally identical with the same objects, method signatures and return values.
"""
# Disable missing docstrings, these are inherited from the async version of the objects
# ruff: noqa: D102
from __future__ import annotations

from collections.abc import Generator
from functools import partial, update_wrapper
from typing import cast

from . import asyncio, objects, portforward
from ._api import ALL
from ._api import Api as _AsyncApi
from ._async_utils import as_sync_func as _as_sync_func
from ._async_utils import as_sync_generator as _as_sync_generator
from ._exceptions import (
    APITimeoutError,
    ConnectionClosedError,
    ExecError,
    NotFoundError,
    ServerError,
)
from ._objects import APIObject
from .asyncio import (
    api as _api,
)
from .asyncio import (
    api_resources as _api_resources,
)
from .asyncio import (
    get as _get,
)
from .asyncio import (
    version as _k8s_version,
)
from .asyncio import (
    watch as _watch,
)
from .asyncio import (
    whoami as _whoami,
)

try:
    from ._version import version as __version__  # noqa
    from ._version import version_tuple as __version_tuple__  # noqa
except ImportError:
    __version__ = "0.0.0"
    __version_tuple__ = (0, 0, 0)


class Api(_AsyncApi):
    _asyncio = False

    def version(self) -> dict:  # type: ignore[override]
        return _as_sync_func(self.async_version)()

    def reauthenticate(self):  # type: ignore[override]
        return _as_sync_func(self.async_reauthenticate)()

    def whoami(self):  # type: ignore[override]
        return _as_sync_func(self.async_whoami)()

    def lookup_kind(self, kind) -> tuple[str, str, bool]:  # type: ignore[override]
        return _as_sync_func(self.async_lookup_kind)(kind)

    def get(  # type: ignore[override]
        self,
        kind: str | type,
        *names: str,
        namespace: str | None = None,
        label_selector: str | dict | None = None,
        field_selector: str | dict | None = None,
        as_object: type[APIObject] | None = None,
        allow_unknown_type: bool = True,
        **kwargs,
    ) -> Generator[objects.APIObject]:
        yield from cast(
            Generator[objects.APIObject],
            _as_sync_generator(self.async_get)(
                kind,
                *names,
                namespace=namespace,
                label_selector=label_selector,
                field_selector=field_selector,
                as_object=as_object,
                allow_unknown_type=allow_unknown_type,
                **kwargs,
            ),
        )

    def watch(  # type: ignore[override]
        self,
        kind: str,
        namespace: str | None = None,
        label_selector: str | dict | None = None,
        field_selector: str | dict | None = None,
        since: str | None = None,
    ) -> Generator[tuple[str, objects.APIObject]]:
        yield from cast(
            Generator[tuple[str, objects.APIObject]],
            _as_sync_generator(self.async_watch)(
                kind,
                namespace=namespace,
                label_selector=label_selector,
                field_selector=field_selector,
                since=since,
            ),
        )

    def api_resources(self) -> list[dict]:  # type: ignore[override]
        return _as_sync_func(self.async_api_resources)()

    def api_versions(self) -> Generator[str]:  # type: ignore[override]
        yield from _as_sync_generator(self.async_api_versions)()

    def create(self, resources: list[objects.APIObject]):  # type: ignore[override]
        return _as_sync_func(self.async_create)(
            cast(list[asyncio.objects.APIObject], resources)
        )


def get(
    kind: str,
    *names: str,
    namespace: str | None = None,
    label_selector: str | dict | None = None,
    field_selector: str | dict | None = None,
    as_object: type | None = None,
    allow_unknown_type: bool = True,
    api=None,
    **kwargs,
):
    """Get a resource by name.

    Args:
        kind: The kind of resource to get
        *names: The names of the resources to get
        namespace: The namespace to get the resource from
        label_selector: The label selector to filter the resources by
        field_selector: The field selector to filter the resources by
        as_object: The object to populate with the resource data
        allow_unknown_type: Whether to allow unknown types
        api: The api to use to get the resource
        **kwargs: Additional arguments to pass to the API

    Returns:
        The populated object

    Raises:
        ValueError: If the resource is not found

    Examples:
        >>> import kr8s
        >>> # All of these are equivalent
        >>> ings = kr8s.get("ing")                           # Short name
        >>> ings = kr8s.get("ingress")                       # Singular
        >>> ings = kr8s.get("ingresses")                     # Plural
        >>> ings = kr8s.get("Ingress")                       # Title
        >>> ings = kr8s.get("ingress.networking.k8s.io")     # Full group name
        >>> ings = kr8s.get("ingress.v1.networking.k8s.io")  # Full with explicit version
        >>> ings = kr8s.get("ingress.networking.k8s.io/v1")  # Full with explicit version alt.
    """
    return _as_sync_generator(_get)(
        kind,
        *names,
        namespace=namespace,
        label_selector=label_selector,
        field_selector=field_selector,
        as_object=as_object,
        allow_unknown_type=allow_unknown_type,
        api=api,
        _asyncio=False,
        **kwargs,
    )


def api(
    url: str | None = None,
    kubeconfig: str | None = None,
    serviceaccount: str | None = None,
    namespace: str | None = None,
    context: str | None = None,
) -> Api:
    """Create a :class:`kr8s.Api` object for interacting with the Kubernetes API.

    If a kr8s object already exists with the same arguments in this thread, it will be returned.

    Args:
        url: The URL of the Kubernetes API server
        kubeconfig: The path to a kubeconfig file to use
        serviceaccount: The path of a service account to use
        namespace: The namespace to use
        context: The context to use

    Returns:
        The API object

    Examples:
        >>> import kr8s
        >>> api = kr8s.api()  # Uses the default kubeconfig
        >>> print(api.version())  # Get the Kubernetes version
    """
    ret = _as_sync_func(_api)(
        url=url,
        kubeconfig=kubeconfig,
        serviceaccount=serviceaccount,
        namespace=namespace,
        context=context,
        _asyncio=False,
    )
    assert isinstance(ret, Api)
    return ret


def whoami():
    """Get the current user's identity.

    Returns:
        The user's identity

    Examples:
        >>> import kr8s
        >>> print(kr8s.whoami())
    """
    return _as_sync_func(_whoami)(_asyncio=False)


def create(resources: list[type[APIObject]], api=None):
    """Creates resources in the Kubernetes cluster."""
    if api is None:
        api = _as_sync_func(_api)(_asyncio=False)
    api.create(cast(list[asyncio.objects.APIObject], resources))


version = _as_sync_func(partial(_k8s_version, _asyncio=False))
update_wrapper(version, _k8s_version)
watch = _as_sync_generator(partial(_watch, _asyncio=False))
update_wrapper(watch, _watch)
api_resources = _as_sync_func(partial(_api_resources, _asyncio=False))
update_wrapper(api_resources, _api_resources)

__all__ = [
    "__version__",
    "__version_tuple__",
    "ALL",
    "api",
    "api_resources",
    "asyncio",
    "create",
    "get",
    "objects",
    "portforward",
    "version",
    "watch",
    "whoami",
    "Api",
    "APITimeoutError",
    "ConnectionClosedError",
    "ExecError",
    "NotFoundError",
    "ServerError",
]



# SPDX-FileCopyrightText: Copyright (c) 2024-2025, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
from kubectl_ng.cli import app
from typer.testing import CliRunner

runner = CliRunner()


def test_get_pods():
    result = runner.invoke(app, ["get", "pods", "-A"])
    assert result.exit_code == 0



# SPDX-FileCopyrightText: Copyright (c) 2023-2025, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
"""Objects to represent Kubernetes resources.

This module provides classes that represent Kubernetes resources.
These classes are used to interact with resources in the Kubernetes API server.
"""
from kr8s._objects import (
    APIObject,
    Binding,
    ClusterRole,
    ClusterRoleBinding,
    ComponentStatus,
    ConfigMap,
    ControllerRevision,
    CronJob,
    CustomResourceDefinition,
    DaemonSet,
    Deployment,
    Endpoints,
    Event,
    HorizontalPodAutoscaler,
    Ingress,
    IngressClass,
    IPAddress,
    Job,
    LimitRange,
    Namespace,
    NetworkPolicy,
    Node,
    PersistentVolume,
    PersistentVolumeClaim,
    Pod,
    PodDisruptionBudget,
    PodTemplate,
    ReplicaSet,
    ReplicationController,
    ResourceQuota,
    Role,
    RoleBinding,
    Secret,
    Service,
    ServiceAccount,
    ServiceCIDR,
    StatefulSet,
    Table,
    get_class,
    new_class,
    object_from_name_type,
    object_from_spec,
    objects_from_files,
)

__all__ = [
    "APIObject",
    "Binding",
    "ClusterRole",
    "ClusterRoleBinding",
    "ComponentStatus",
    "ConfigMap",
    "ControllerRevision",
    "CronJob",
    "CustomResourceDefinition",
    "DaemonSet",
    "Deployment",
    "Endpoints",
    "Event",
    "HorizontalPodAutoscaler",
    "Ingress",
    "IngressClass",
    "Job",
    "LimitRange",
    "Namespace",
    "NetworkPolicy",
    "Node",
    "PersistentVolume",
    "PersistentVolumeClaim",
    "Pod",
    "PodDisruptionBudget",
    "PodTemplate",
    "ReplicaSet",
    "ReplicationController",
    "ResourceQuota",
    "Role",
    "RoleBinding",
    "Secret",
    "Service",
    "ServiceAccount",
    "StatefulSet",
    "Table",
    "IPAddress",
    "ServiceCIDR",
    "object_from_name_type",
    "objects_from_files",
    "get_class",
    "new_class",
    "object_from_spec",
]



# SPDX-FileCopyrightText: Copyright (c) 2023-2025, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
from __future__ import annotations

import asyncio
import contextlib
import random
import socket
import sys
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager, suppress
from typing import TYPE_CHECKING, Literal, Union

import anyio
import httpx_ws
import sniffio

from ._exceptions import ConnectionClosedError
from ._types import APIObjectWithPods

LocalPortType = Union[Literal["match", "auto"], int, None]

if TYPE_CHECKING:
    from ._objects import APIObject

if sys.version_info < (3, 12, 1):
    # contextlib.supress() in Python 3.12.1 supprts ExceptionGroups
    # For older versions, we use the exceptiongroup backport
    from exceptiongroup import suppress  # type: ignore # noqa: F811


class PortForward:
    """Start a tcp server and forward all connections to a Pod port.

    You can either pass a :class:`kr8s.objects.Pod` or any resource with a ``ready_pods`` method
    such as a :class:`kr8s.objects.Service`.

    .. note::
        The ``ready_pods`` method should return a list of Pods that are ready to accept connections.

    .. warning:
        Currently Port Forwards only work when using ``asyncio`` and not ``trio``.

    Args:
        resource:
            The Pod or Resource to forward to.
        remote_port:
            The port on the Pod to forward to.
        local_port:
            The local port to listen on. Defaults to ``"match"``, which will match the ``remote_port``.
            Set to ``"auto"`` or ``None`` to find an available high port.
            Set to an ``int`` to specify a specific port.
        address:
            List of addresses or address to listen on. Defaults to ["127.0.0.1"], will listen only on 127.0.0.1.

    Example:
        This class can be used as a an async context manager or with explicit start/stop methods.

        Context manager:

        >>> async with PortForward(pod, 8888) as port:
        ...     print(f"Forwarding to port {port}")
        ...     # Do something with port 8888 on the Pod


        Explict start/stop:

        >>> pf = PortForward(pod, 8888)
        >>> await pf.start()
        >>> print(f"Forwarding to port {pf.local_port}")
        >>> # Do something with port 8888 on the Pod
        >>> await pf.stop()

        Explict bind address:

        >>> async with PortForward(pod, 8888, address=["127.0.0.1", "10.20.0.1"]) as port:
        ...     print(f"Forwarding to port {port}")
        ...     # Do something with port 8888 on the Pod, port will be bind to 127.0.0.1 and 10.20.0.1

    """

    def __init__(
        self,
        resource: APIObject,
        remote_port: int,
        local_port: LocalPortType = "match",
        address: list[str] | str = "127.0.0.1",
    ) -> None:
        with suppress(sniffio.AsyncLibraryNotFoundError):
            if sniffio.current_async_library() != "asyncio":
                raise RuntimeError(
                    "PortForward only works with asyncio, "
                    "see https://github.com/kr8s-org/kr8s/issues/104"
                )
        self.server = None
        self.servers: list[asyncio.Server] = []
        self.remote_port = remote_port
        if local_port == "match":
            self.local_port = remote_port
        elif local_port == "auto" or local_port is None:
            self.local_port = 0
        elif isinstance(local_port, int):
            self.local_port = local_port
        else:
            raise TypeError("local_port must be 'match', 'auto', an int or None")
        if isinstance(address, str):
            address = [address]
        self.address = address
        from ._objects import Pod

        if not isinstance(resource, Pod) and not hasattr(resource, "ready_pods"):
            raise ValueError(
                "resource must be a Pod or a resource with a ready_pods method"
            )
        self._resource = resource
        self.pod = None
        self._loop: asyncio.AbstractEventLoop | None = None
        self._tasks: list[asyncio.Task] = []
        self._run_task = None
        self._bg_future: asyncio.Future | None = None
        self._bg_task: asyncio.Task | None = None

    async def __aenter__(self, *args, **kwargs):
        self._run_task = self._run()
        return await self._run_task.__aenter__(*args, **kwargs)

    async def __aexit__(self, *args, **kwargs):
        assert self._run_task
        return await self._run_task.__aexit__(*args, **kwargs)

    async def start(self) -> int:
        """Start a background task with the port forward running."""
        return await self.async_start()

    async def async_start(self) -> int:
        if self._loop is None:
            self._loop = asyncio.get_event_loop()
        if self._bg_task is not None:
            return self.local_port

        async def f():
            self._bg_future = self._loop.create_future()
            async with self as port:
                self.local_port = port
                await self._bg_future

        self._bg_task = self._loop.create_task(f())
        while self.local_port == 0:
            await anyio.sleep(0.1)
        return self.local_port

    async def stop(self) -> None:
        """Stop the background task."""
        return await self.async_stop()

    async def async_stop(self) -> None:
        if self._bg_future:
            self._bg_future.set_result(None)
        self._bg_task = None

    async def run_forever(self) -> None:
        """Run the port forward forever.

        Example:
            >>> pf = pod.portforward(remote_port=8888, local_port=8889)
            >>> # or
            >>> pf = PortForward(pod, remote_port=8888, local_port=8889)
            >>> await pf.run_forever()
        """
        return await self.async_run_forever()

    async def async_run_forever(self) -> None:
        async with self:
            with contextlib.suppress(asyncio.CancelledError):
                for server in self.servers:
                    await server.serve_forever()

    @asynccontextmanager
    async def _run(self) -> AsyncGenerator[int]:
        """Start the port forward for multiple bind addresses and yield the local port."""
        if self.local_port == 0:
            self.local_port = self._find_available_port()

        for address in self.address:
            server = await asyncio.start_server(
                self._sync_sockets, port=self.local_port, host=address
            )
            self.servers.append(server)

        try:
            for server in self.servers:
                await server.start_serving()
            yield self.local_port

        finally:
            # Ensure all servers are closed properly
            for server in self.servers:
                server.close()
                await server.wait_closed()
                self.servers.remove(server)

    async def _select_pod(self) -> APIObject:
        """Select a Pod to forward to."""
        from ._objects import Pod

        if isinstance(self._resource, Pod):
            return self._resource

        if isinstance(self._resource, APIObjectWithPods):
            try:
                return random.choice(await self._resource.async_ready_pods())
            except IndexError:
                pass
        raise RuntimeError("No ready pods found")

    @asynccontextmanager
    async def _connect_websocket(self):
        """Connect to the Kubernetes portforward websocket."""
        connection_attempts = 0
        while True:
            if not self.pod:
                self.pod = await self._select_pod()
            try:
                assert self.pod.api
                async with self.pod.api.open_websocket(
                    version=self.pod.version,
                    url=f"{self.pod.endpoint}/{self.pod.name}/portforward",
                    namespace=self.pod.namespace,
                    params={
                        "name": self.pod.name,
                        "namespace": self.pod.namespace,
                        "ports": f"{self.remote_port}",
                        "_preload_content": "false",
                    },
                ) as websocket:
                    yield websocket
                    break
            except httpx_ws.HTTPXWSException as e:
                self.pod = None
                if connection_attempts > 5:
                    raise ConnectionClosedError("Unable to connect to Pod") from e
                await anyio.sleep(0.1 * connection_attempts)

    async def _sync_sockets(self, reader, writer) -> None:
        """Start two tasks to copy bytes from tcp=>websocket and websocket=>tcp."""
        try:
            async with self._connect_websocket() as ws:
                with suppress(ConnectionClosedError, httpx_ws.WebSocketDisconnect):
                    async with anyio.create_task_group() as tg:
                        tg.start_soon(self._tcp_to_ws, ws, reader)
                        tg.start_soon(self._ws_to_tcp, ws, writer)
        finally:
            writer.close()

    async def _tcp_to_ws(self, ws, reader) -> None:
        while True:
            try:
                data = await reader.read(1024 * 1024)
            except ConnectionResetError:
                data = None
            if not data:
                raise ConnectionClosedError("TCP socket closed")
            else:
                # Send data to channel 0 of the websocket.
                # TODO Support multiple channels for multiple ports.
                try:
                    await ws.send_bytes(b"\x00" + data)
                except ConnectionResetError as e:
                    raise ConnectionClosedError("Websocket closed") from e

    async def _ws_to_tcp(self, ws, writer) -> None:
        channels = []
        while True:
            message = await ws.receive_bytes()
            # Kubernetes portforward protocol prefixes all frames with a byte to represent
            # the channel. Channel 0 is rw for data and channel 1 is ro for errors.
            if message[0] not in channels:
                # Keep track of our channels. Could be useful later for listening to multiple ports.
                channels.append(message[0])
            else:
                if message[0] % 2 == 1:  # pragma: no cover
                    # Odd channels are for errors.
                    raise ConnectionClosedError(message[1:].decode())
                writer.write(message[1:])
                await writer.drain()

    def _is_port_in_use(self, port: int, host: str = "127.0.0.1") -> bool:
        """Check if a given port is in use on a specified host.

        Args:
            port: Port number to check.
            host: Host address to check the port on. Default is localhost.

        Returns:
            bool: True if the port is in use, False otherwise.
        """
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            return s.connect_ex((host, port)) == 0

    def _find_available_port(self):
        """Find a random port that is not in use on any of the given addresses.

        Returns:
            An available port number.
        """
        while True:
            port = random.randint(10000, 60000)
            if not any(self._is_port_in_use(port, address) for address in self.address):
                return port



#!/usr/bin/env python
# SPDX-FileCopyrightText: Copyright (c) 2025, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
# /// script
# requires-python = ">=3.12"
# dependencies = [
#     "ruamel-yaml",
# ]
# ///
import json
import os
import re
import urllib.request
from datetime import datetime
from pathlib import Path

from ruamel.yaml import YAML

yaml = YAML(typ="rt")
yaml.width = 2**16
yaml.preserve_quotes = True
yaml.indent(mapping=2, sequence=4, offset=2)

DATE_FORMAT = "%Y-%m-%d"
SUPPORT_MODE = "extended"  # "standard" or "extended"


def extract_dates(data):
    if "extendedSupport" in data and isinstance(data["extendedSupport"], str):
        yield datetime.strptime(data["extendedSupport"], DATE_FORMAT)
    if "eol" in data and isinstance(data["eol"], str):
        yield datetime.strptime(data["eol"], DATE_FORMAT)
    if "lts" in data and isinstance(data["lts"], str):
        yield datetime.strptime(data["lts"], DATE_FORMAT)


def has_eol(data):
    """Check if a version has a defined end of life."""
    if "eol" in data and isinstance(data["eol"], str):
        return True
    if "extendedSupport" in data and isinstance(data["extendedSupport"], str):
        return True
    if "lts" in data and isinstance(data["lts"], str):
        return True
    return False


def get_support_date(data):
    dates = list(extract_dates(data))
    if not dates:
        return None
    if SUPPORT_MODE == "standard":
        return min(dates)
    elif SUPPORT_MODE == "extended":
        return max(dates)


def get_kubernetes_oss_versions():
    print(
        "Loading Kubernetes versions from https://endoflife.date/api/kubernetes.json..."
    )
    with urllib.request.urlopen("https://endoflife.date/api/kubernetes.json") as url:
        data = json.load(url)
        data = [
            {
                "cycle": x["cycle"],
                "latest_version": x["latest"],
                "eol": get_support_date(x),
            }
            for x in data
            if has_eol(x)
        ]
        data.sort(key=lambda x: x["eol"], reverse=True)
    return data


def get_azure_aks_versions():
    url = "https://endoflife.date/api/azure-kubernetes-service.json"
    print(f"Loading Azure AKS versions from {url}...")
    with urllib.request.urlopen(url) as payload:
        data = json.load(payload)

        data = [
            {
                "cycle": x["cycle"],
                "eol": get_support_date(x),
            }
            for x in data
            if has_eol(x)
        ]
        data.sort(key=lambda x: x["eol"], reverse=True)
    return data


def get_amazon_eks_versions():
    url = "https://endoflife.date/api/amazon-eks.json"
    print(f"Loading Amazon EKS versions from {url}...")
    with urllib.request.urlopen(url) as payload:
        data = json.load(payload)
        data = [
            {
                "cycle": x["cycle"],
                "eol": get_support_date(x),
            }
            for x in data
            if has_eol(x)
        ]
        data.sort(key=lambda x: x["eol"], reverse=True)
    return data


def get_google_kubernetes_engine_versions():
    url = "https://endoflife.date/api/google-kubernetes-engine.json"
    print(f"Loading Google Kubernetes Engine versions from {url}...")
    with urllib.request.urlopen(url) as payload:
        data = json.load(payload)
        data = [
            {
                "cycle": x["cycle"],
                "eol": get_support_date(x),
            }
            for x in data
            if has_eol(x)
        ]
        data.sort(key=lambda x: x["eol"], reverse=True)
    return data


def extend_versions(versions, extended_versions, provider):
    print(f"Extending EOL dates with {provider} support dates...")
    for extended_version in extended_versions:
        for version in versions:
            if version["cycle"] == extended_version["cycle"]:
                if version["eol"] < extended_version["eol"]:
                    print(
                        f"Extending EOL date for {version['cycle']} from {version['eol']:%Y-%m-%d} to "
                        f"{provider} support date {extended_version['eol']:%Y-%m-%d}"
                    )
                    version["eol"] = extended_version["eol"]
    return versions


def get_kind_versions():
    print("Loading Kubernetes tags from https://hub.docker.com/r/kindest/node/tags...")
    container_tags = []
    next_url = "https://hub.docker.com/v2/repositories/kindest/node/tags"
    while next_url:
        with urllib.request.urlopen(next_url) as url:
            results = json.load(url)
            container_tags += results["results"]
            if "next" in results and results["next"]:
                next_url = results["next"]
            else:
                next_url = None
    return container_tags


def get_versions():
    versions = get_kubernetes_oss_versions()
    versions = extend_versions(versions, get_azure_aks_versions(), "Azure AKS")
    versions = extend_versions(versions, get_amazon_eks_versions(), "Amazon EKS")
    versions = extend_versions(
        versions, get_google_kubernetes_engine_versions(), "Google Kubernetes Engine"
    )
    print("Pruning versions that are past their support date...")
    versions = [x for x in versions if x["eol"] > datetime.now()]

    container_tags = get_kind_versions()

    for version in versions:
        try:
            version["latest_kind_container"] = [
                x["name"]
                for x in container_tags
                if version["cycle"] in x["name"] and "alpha" not in x["name"]
            ][0][1:]
        except IndexError:
            version["latest_kind_container"] = None

    before_length = len(versions)
    print("Pruning versions that do not have a kind release yet...")
    versions[:] = [x for x in versions if x["latest_kind_container"] is not None]
    print(f"Pruned {before_length - len(versions)} versions")
    return versions


def update_workflow(versions, workflow_path):
    workflow_path = Path(workflow_path)
    workflow = yaml.load(workflow_path)
    latest_kind_container = versions[0]["latest_kind_container"]
    if "minimal-deps" in workflow["jobs"]:
        workflow["jobs"]["minimal-deps"]["strategy"]["matrix"]["kubernetes-version"][
            0
        ] = latest_kind_container
    workflow["jobs"]["test"]["strategy"]["matrix"]["kubernetes-version"][
        0
    ] = latest_kind_container
    workflow["jobs"]["test"]["strategy"]["matrix"]["include"] = []
    for version in versions[1:]:
        workflow["jobs"]["test"]["strategy"]["matrix"]["include"].append(
            {
                "python-version": "3.10",
                "kubernetes-version": version["latest_kind_container"],
            }
        )
    yaml.dump(workflow, workflow_path)


def update_badges(filename, versions):
    readme = Path(filename).read_text()
    # Use regex to replace the badge
    v = [x["cycle"] for x in versions]
    v.sort()
    version_list = "%7C".join(v)
    readme = re.sub(
        r"img.shields.io/badge/Kubernetes%20support.*-blue",
        f"img.shields.io/badge/Kubernetes%20support-{version_list}-blue",
        readme,
    )
    Path(filename).write_text(readme)


def main():
    versions = get_versions()
    print(f"Latest version: {versions[0]['cycle']}")
    print("Supported versions:")
    for version in versions:
        print(
            f"For {version['cycle']} using kindest/node {version['latest_kind_container']}"
            f" until {version['eol']:%Y-%m-%d}"
        )

    if not os.environ.get("DEBUG"):
        update_workflow(versions, ".github/workflows/test-kr8s.yaml")
        update_workflow(versions, ".github/workflows/test-kubectl-ng.yaml")
        update_badges("README.md", versions)
        update_badges("docs/index.md", versions)
    else:
        print("DEBUG env var set, skipping file updates")


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: Copyright (c) 2024-2025, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
"""Objects to represent Kubernetes resources.

This module provides classes that represent Kubernetes resources.
These classes are used to interact with resources in the Kubernetes API server.
"""

# Disable missing docstrings, these are inherited from the async version of the objects
# ruff: noqa: D102
from __future__ import annotations

import pathlib
from collections.abc import Generator
from typing import TYPE_CHECKING, Any, BinaryIO, cast

import httpx

from ._async_utils import as_sync_func, as_sync_generator
from ._exec import CompletedExec
from ._objects import APIObjectSyncMixin
from ._objects import (
    Binding as _Binding,
)
from ._objects import (
    ClusterRole as _ClusterRole,
)
from ._objects import (
    ClusterRoleBinding as _ClusterRoleBinding,
)
from ._objects import (
    ComponentStatus as _ComponentStatus,
)
from ._objects import (
    ConfigMap as _ConfigMap,
)
from ._objects import (
    ControllerRevision as _ControllerRevision,
)
from ._objects import (
    CronJob as _CronJob,
)
from ._objects import (
    CustomResourceDefinition as _CustomResourceDefinition,
)
from ._objects import (
    DaemonSet as _DaemonSet,
)
from ._objects import (
    Deployment as _Deployment,
)
from ._objects import (
    Endpoints as _Endpoints,
)
from ._objects import (
    Event as _Event,
)
from ._objects import (
    HorizontalPodAutoscaler as _HorizontalPodAutoscaler,
)
from ._objects import (
    Ingress as _Ingress,
)
from ._objects import (
    IngressClass as _IngressClass,
)
from ._objects import (
    IPAddress as _IPAddress,
)
from ._objects import (
    Job as _Job,
)
from ._objects import (
    LimitRange as _LimitRange,
)
from ._objects import (
    Namespace as _Namespace,
)
from ._objects import (
    NetworkPolicy as _NetworkPolicy,
)
from ._objects import (
    Node as _Node,
)
from ._objects import (
    PersistentVolume as _PersistentVolume,
)
from ._objects import (
    PersistentVolumeClaim as _PersistentVolumeClaim,
)
from ._objects import (
    Pod as _Pod,
)
from ._objects import (
    PodDisruptionBudget as _PodDisruptionBudget,
)
from ._objects import (
    PodTemplate as _PodTemplate,
)
from ._objects import (
    ReplicaSet as _ReplicaSet,
)
from ._objects import (
    ReplicationController as _ReplicationController,
)
from ._objects import (
    ResourceQuota as _ResourceQuota,
)
from ._objects import (
    Role as _Role,
)
from ._objects import (
    RoleBinding as _RoleBinding,
)
from ._objects import (
    Secret as _Secret,
)
from ._objects import (
    Service as _Service,
)
from ._objects import (
    ServiceAccount as _ServiceAccount,
)
from ._objects import (
    ServiceCIDR as _ServiceCIDR,
)
from ._objects import (
    StatefulSet as _StatefulSet,
)
from ._objects import (
    Table as _Table,
)
from ._objects import (
    get_class as _get_class,
)
from ._objects import (
    new_class as _new_class,
)

# noqa
from ._objects import object_from_name_type as _object_from_name_type
from ._objects import (
    object_from_spec as _object_from_spec,
)
from ._objects import objects_from_files as _objects_from_files
from .portforward import PortForward

if TYPE_CHECKING:
    from kr8s import Api


class APIObject(APIObjectSyncMixin):
    __doc__ = APIObjectSyncMixin.__doc__


class Binding(APIObjectSyncMixin, _Binding):
    __doc__ = _Binding.__doc__


class ComponentStatus(APIObjectSyncMixin, _ComponentStatus):
    __doc__ = _ComponentStatus.__doc__


class ConfigMap(APIObjectSyncMixin, _ConfigMap):
    __doc__ = _ConfigMap.__doc__


class Endpoints(APIObjectSyncMixin, _Endpoints):
    __doc__ = _Endpoints.__doc__


class Event(APIObjectSyncMixin, _Event):
    __doc__ = _Event.__doc__


class LimitRange(APIObjectSyncMixin, _LimitRange):
    __doc__ = _LimitRange.__doc__


class Namespace(APIObjectSyncMixin, _Namespace):
    __doc__ = _Namespace.__doc__


class Node(APIObjectSyncMixin, _Node):

    def cordon(self):
        return as_sync_func(self.async_cordon)()

    def uncordon(self):
        return as_sync_func(self.async_uncordon)()

    def taint(self, key, value, *, effect):
        return as_sync_func(self.async_taint)(key, value, effect=effect)


class PersistentVolume(APIObjectSyncMixin, _PersistentVolume):
    __doc__ = _PersistentVolume.__doc__


class PersistentVolumeClaim(APIObjectSyncMixin, _PersistentVolumeClaim):
    __doc__ = _PersistentVolumeClaim.__doc__


class Pod(APIObjectSyncMixin, _Pod):
    def ready(self) -> bool:  # type: ignore[override]
        return as_sync_func(self.async_ready)()

    def logs(  # type: ignore[override]
        self,
        container=None,
        pretty=None,
        previous=False,
        since_seconds=None,
        since_time=None,
        timestamps=False,
        tail_lines=None,
        limit_bytes=None,
        follow=False,
        timeout=3600,
    ) -> Generator[str]:
        yield from as_sync_generator(self.async_logs)(
            container,
            pretty,
            previous,
            since_seconds,
            since_time,
            timestamps,
            tail_lines,
            limit_bytes,
            follow,
            timeout,
        )

    def exec(  # type: ignore[override]
        self,
        command: list[str],
        *,
        container: str | None = None,
        stdin: str | BinaryIO | None = None,
        stdout: BinaryIO | None = None,
        stderr: BinaryIO | None = None,
        check: bool = True,
        capture_output: bool = True,
    ) -> CompletedExec:
        return as_sync_func(self.async_exec)(
            command,
            container=container,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            check=check,
            capture_output=capture_output,
        )

    def tolerate(  # type: ignore[override]
        self,
        key: str,
        *,
        operator: str,
        effect: str,
        value: str | None = None,
        toleration_seconds: int | None = None,
    ):
        return as_sync_func(self.async_tolerate)(
            key,
            operator=operator,
            effect=effect,
            value=value,
            toleration_seconds=toleration_seconds,
        )

    def portforward(
        self, remote_port, local_port="match", address="127.0.0.1"
    ) -> PortForward:
        pf = super().portforward(remote_port, local_port, address)
        assert isinstance(pf, PortForward)
        return pf


class PodTemplate(APIObjectSyncMixin, _PodTemplate):
    __doc__ = _PodTemplate.__doc__


class ReplicationController(APIObjectSyncMixin, _ReplicationController):

    def ready(self):
        return as_sync_func(self.async_ready)()


class ResourceQuota(APIObjectSyncMixin, _ResourceQuota):
    __doc__ = _ResourceQuota.__doc__


class Secret(APIObjectSyncMixin, _Secret):
    __doc__ = _Secret.__doc__


class ServiceAccount(APIObjectSyncMixin, _ServiceAccount):
    __doc__ = _ServiceAccount.__doc__


class Service(APIObjectSyncMixin, _Service):
    def proxy_http_request(  # type: ignore[override]
        self, method: str, path: str, port: int | None = None, **kwargs: Any
    ) -> httpx.Response:
        return as_sync_func(self.async_proxy_http_request)(
            method, path, port=port, **kwargs
        )

    def proxy_http_get(  # type: ignore[override]
        self, path: str, port: int | None = None, **kwargs
    ) -> httpx.Response:
        return as_sync_func(self.async_proxy_http_request)("GET", path, port, **kwargs)

    def proxy_http_post(  # type: ignore[override]
        self, path: str, port: int | None = None, **kwargs
    ) -> httpx.Response:
        return as_sync_func(self.async_proxy_http_request)("POST", path, port, **kwargs)

    def proxy_http_put(  # type: ignore[override]
        self, path: str, port: int | None = None, **kwargs
    ) -> httpx.Response:
        return as_sync_func(self.async_proxy_http_request)("PUT", path, port, **kwargs)

    def proxy_http_delete(  # type: ignore[override]
        self, path: str, port: int | None = None, **kwargs
    ) -> httpx.Response:
        return as_sync_func(self.async_proxy_http_request)(
            "DELETE", path, port, **kwargs
        )

    def ready_pods(self) -> list[Pod]:  # type: ignore[override]
        return cast(list[Pod], as_sync_func(self.async_ready_pods)())

    def ready(self):
        return as_sync_func(self.async_ready)()

    def portforward(
        self, remote_port, local_port="match", address="127.0.0.1"
    ) -> PortForward:
        pf = super().portforward(remote_port, local_port, address)
        assert isinstance(pf, PortForward)
        return pf


class ControllerRevision(APIObjectSyncMixin, _ControllerRevision):
    __doc__ = _ControllerRevision.__doc__


class DaemonSet(APIObjectSyncMixin, _DaemonSet):
    __doc__ = _DaemonSet.__doc__


class Deployment(APIObjectSyncMixin, _Deployment):

    def pods(self) -> list[Pod]:  # type: ignore[override]
        return cast(list[Pod], as_sync_func(self.async_pods)())

    def ready(self) -> bool:  # type: ignore[override]
        return as_sync_func(self.async_ready)()


class ReplicaSet(APIObjectSyncMixin, _ReplicaSet):
    __doc__ = _ReplicaSet.__doc__


class StatefulSet(APIObjectSyncMixin, _StatefulSet):
    __doc__ = _StatefulSet.__doc__


class HorizontalPodAutoscaler(APIObjectSyncMixin, _HorizontalPodAutoscaler):
    __doc__ = _HorizontalPodAutoscaler.__doc__


class CronJob(APIObjectSyncMixin, _CronJob):
    __doc__ = _CronJob.__doc__


class Job(APIObjectSyncMixin, _Job):
    __doc__ = _Job.__doc__


class Ingress(APIObjectSyncMixin, _Ingress):
    __doc__ = _Ingress.__doc__


class IngressClass(APIObjectSyncMixin, _IngressClass):
    __doc__ = _IngressClass.__doc__


class NetworkPolicy(APIObjectSyncMixin, _NetworkPolicy):
    __doc__ = _NetworkPolicy.__doc__


class PodDisruptionBudget(APIObjectSyncMixin, _PodDisruptionBudget):
    __doc__ = _PodDisruptionBudget.__doc__


class ClusterRoleBinding(APIObjectSyncMixin, _ClusterRoleBinding):
    __doc__ = _ClusterRoleBinding.__doc__


class ClusterRole(APIObjectSyncMixin, _ClusterRole):
    __doc__ = _ClusterRole.__doc__


class RoleBinding(APIObjectSyncMixin, _RoleBinding):
    __doc__ = _RoleBinding.__doc__


class Role(APIObjectSyncMixin, _Role):
    __doc__ = _Role.__doc__


class CustomResourceDefinition(APIObjectSyncMixin, _CustomResourceDefinition):
    __doc__ = _CustomResourceDefinition.__doc__


class Table(APIObjectSyncMixin, _Table):
    __doc__ = _Table.__doc__


class IPAddress(APIObjectSyncMixin, _IPAddress):
    __doc__ = _IPAddress.__doc__


class ServiceCIDR(APIObjectSyncMixin, _ServiceCIDR):
    __doc__ = _ServiceCIDR.__doc__


def new_class(
    kind: str,
    version: str | None = None,
    asyncio: bool = False,
    namespaced=True,
    scalable: bool | None = None,
    scalable_spec: str | None = None,
    plural: str | None = None,
) -> type[APIObject]:
    """Create a new APIObject subclass.

    Args:
        kind: The Kubernetes resource kind.
        version: The Kubernetes API version.
        asyncio: Whether to use asyncio or not.
        namespaced: Whether the resource is namespaced or not.
        scalable: Whether the resource is scalable or not.
        scalable_spec: The name of the field to use for scaling.
        plural: The plural form of the resource.

    Returns:
        A new APIObject subclass.
    """
    return cast(
        type[APIObject],
        _new_class(kind, version, asyncio, namespaced, scalable, scalable_spec, plural),
    )


def object_from_name_type(
    name: str,
    namespace: str | None = None,
    api: Api | None = None,
) -> APIObject:
    """Create an APIObject from a Kubernetes resource name.

    Args:
        name: A Kubernetes resource name.
        namespace: The namespace of the resource.
        api: An optional API instance to use.

    Returns:
        A corresponding APIObject subclass instance.

    Raises:
        ValueError: If the resource kind or API version is not supported.
    """
    return cast(
        APIObject,
        as_sync_func(_object_from_name_type)(name, namespace, api, _asyncio=False),
    )


def objects_from_files(
    path: str | pathlib.Path,
    api: Api | None = None,
    recursive: bool = False,
) -> list[APIObject]:
    """Create APIObjects from Kubernetes resource files.

    Args:
        path: A path to a Kubernetes resource file or directory of resource files.
        api: An optional API instance to use.
        recursive: Whether to recursively search for resource files in subdirectories.

    Returns:
        A list of APIObject subclass instances.

    Raises:
        ValueError: If the resource kind or API version is not supported.
    """
    return cast(
        list[APIObject],
        as_sync_func(_objects_from_files)(path, api, recursive, _asyncio=False),
    )


def get_class(kind: str, version: str) -> type[APIObject]:
    """Get an APIObject subclass by kind and version.

    Args:
        kind: The Kubernetes resource kind.
        version: The Kubernetes API group/version.

    Returns:
        An APIObject subclass.

    Raises:
        KeyError: If no object is registered for the given kind and version.
    """
    return cast(type[APIObject], _get_class(kind, version, _asyncio=False))


def object_from_spec(
    spec: dict, api: Api | None = None, allow_unknown_type: bool = False
) -> APIObject:
    """Create an APIObject from a Kubernetes resource spec.

    Args:
        spec: A Kubernetes resource spec.
        api: An optional API instance to use.
        allow_unknown_type: Whether to allow unknown resource types.

    Returns:
        A corresponding APIObject subclass instance.

    Raises:
        ValueError: If the resource kind or API version is not supported.
    """
    return cast(
        APIObject, _object_from_spec(spec, api, allow_unknown_type, _asyncio=False)
    )



# SPDX-FileCopyrightText: Copyright (c) 2023-2025, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
import time

from kr8s.objects import Pod


def test_gen_pod(ns):
    pod = Pod.gen(
        name="foo",
        namespace=ns,
        image="nginx",
        labels={"app": "foo"},
        annotations={"bar": "baz"},
    )
    assert isinstance(pod, Pod)
    assert pod.name == "foo"
    assert pod.namespace == ns
    assert pod.spec.containers[0].image == "nginx"
    assert "app" in pod.labels
    assert "bar" in pod.annotations
    pod.create()
    while not pod.exists():
        time.sleep(1)
    pod.delete()


def test_gen_simple_pod(ns):
    pod = Pod.gen(name="foo", image="nginx")  # Minimum arguments you can pass
    pod.namespace = ns  # We need to set the namespace to avoid collisions in tests
    pod.create()
    while not pod.exists():
        time.sleep(1)
    pod.delete()



# SPDX-FileCopyrightText: Copyright (c) 2023-2025, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
import os

from kr8s._testutils import set_env


def test_set_env():
    assert "FOO" not in os.environ
    with set_env(FOO="bar"):
        assert "FOO" in os.environ
    assert "FOO" not in os.environ

    os.environ["FOO"] = "bar"
    assert "FOO" in os.environ
    assert os.environ["FOO"] == "bar"
    with set_env(FOO="baz"):
        assert "FOO" in os.environ
        assert os.environ["FOO"] == "baz"
    assert "FOO" in os.environ
    assert os.environ["FOO"] == "bar"
    del os.environ["FOO"]



# SPDX-FileCopyrightText: Copyright (c) 2023-2025, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
import sys

import typer
from rich.console import Console

from kr8s.asyncio.objects import Pod

console = Console()


async def kexec(
    resource: str = typer.Argument(..., help="POD | TYPE/NAME"),
    namespace: str = typer.Option(
        None,
        "-n",
        "--namespace",
    ),
    container: str = typer.Option(
        None,
        "-c",
        "--container",
        help="Container name. "
        "If omitted, use the kubectl.kubernetes.io/default-container annotation "
        "for selecting the container to be attached or the first container in the "
        "pod will be chosen",
    ),
    command: list[str] = typer.Argument(..., help="COMMAND [args...]"),
):
    """Execute a command in a container."""
    pod = await Pod.get(resource, namespace=namespace)
    await pod.exec(
        command,
        container=container,
        stdout=sys.stdout.buffer,
        stderr=sys.stderr.buffer,
        check=False,
    )
